"""Reads a parquet dataset from a directory"""
from typing import Dict, List, Tuple
from pathlib import Path
import numpy as np
import pandas as pd
import pyarrow as pa
from pyarrow import parquet as pq
from loguru import logger
from ..column_types import ColumnType

def _peek_one_file_only(path) -> pa.Schema:
    """
    Parquet exports can also be directories containing multiple singular parquet files.
    pq.read_schema wants only these files, so we provide it the first file inside that directory
    NOTE: this doesn't protect you from invalid parquet files (having different headers) inside the directory
    """
    try:
        schema = pq.read_schema(path)
    except OSError:
        all_files = list(path.iterdir())
        assert len(all_files) > 0, f"Found dir at '{path}' but with no parquet files inside"
        pq_files = [x for x in all_files if x.suffix==".parquet" and x.is_file()]
        assert len(pq_files) == len(all_files), f"Found other type of files at '{path}' except parquet: {all_files}"
        return _peek_one_file_only(pq_files[0])
    return schema

def _pq_to_dataset_type(pq_type) -> ColumnType:
    """Maps parquet types to Dataset types"""
    if pq_type == "string":
        return "str"
    if pq_type in ("float", "double"):
        return "float"
    if pq_type == "bool":
        return "bool"
    if pq_type in ("int64", "int32"):
        return "int"
    if isinstance(pq_type, pa.lib.ListType):
        # Both vectors and objects are stored as lists, so we need to look at the inner type to differnetiate
        if pq_type.value_type == "float":
            return "vector"
        return "object"
    if isinstance(pq_type, pa.lib.DictionaryType):
        return "categorical"
    if isinstance(pq_type, pa.lib.TimestampType):
        assert pq_type.to_pandas_dtype() == np.dtype("datetime64[ns]"), f"Unknown datetime format: {pq_type}"
        return "datetime64[ns]"
    if isinstance(pq_type, pa.lib.DurationType):
        assert pq_type.unit == "ns", f"Unknown timedelta format: {pq_type}"
        return "timedelta64[ns]"
    if isinstance(pq_type, pa.lib.DataType):
        if pq_type == "date32[day]":
            assert pq_type.to_pandas_dtype() == np.dtype("datetime64[ns]"), f"Unknown datetime format: {pq_type}"
            return "datetime64[ns]"

    raise TypeError(f"Unknown parquet type: {pq_type} found in the dataset.")

def parquet_peek(path: Path, tables_pk: Dict[str, str]) -> Dict[str, Dict[str, str]]:
    """Reads just the first row of all pq files given a path. Then, reads the types.json file and returns it."""
    column_types = {}
    for tb, pk in tables_pk.items():
        column_types[tb] = {}
        schema = _peek_one_file_only(path / f"{tb}.parquet")
        assert pk in schema.names, f"Primary key '{tb}/{pk}' not found in the parquet file!"
        for col, pq_type in zip(schema.names, schema.types):
            column_type = _pq_to_dataset_type(pq_type)
            column_types[tb][col] = column_type
    return column_types

def parquet_read(path: Path, tables_pk: Dict[str, str], columns: Dict[str, List[str]]) -> Tuple[Dict, Dict]:
    """Reads data from a path of parquets, given a list of columns and a list of data-groups plus their primary keys"""
    assert path.is_dir(), f"Path must be a directory. Got '{path}'."
    assert isinstance(columns, dict), f"Expected dict, got {columns}"

    logger.info(f"Reading Dataset from parquet files. Path: '{path}'. Data groups: {list(tables_pk)}")
    header = parquet_peek(path, tables_pk)
    data = {}
    dataset_types = {}
    for tb, pk in tables_pk.items():
        file_name = path / f"{tb}.parquet"
        # parquet data type conversion is built-in. No manual needed, like in csv case.
        logger.debug(f"Reading the parquet file: '{file_name}' ({len(columns[tb])} columns)")
        assert columns[tb] is None or isinstance(columns[tb], list), f"Expected list or None, got {columns[tb]}"
        data[tb] = pd.read_parquet(file_name, columns=columns[tb])
        data[tb] = data[tb].set_index(pk) if data[tb].index.name != pk else data[tb]
        dataset_types[tb] = {k: v for k, v in header[tb].items() if k in columns[tb]}
    return data, dataset_types



"""Reads a (Typed) CSV dataset from a directory"""
# pylint: disable=consider-iterating-dictionary
from typing import Dict, List, Callable, Tuple, Optional
from pathlib import Path
import json
import ast
import numpy as np
import pandas as pd
from loguru import logger
from ..column_types import VALID_COLUMN_TYPES, ColumnType, column_type_to_dtype

def _build_converters(dg_column_types: Dict[str, Dict[str, ColumnType]],
                      object_converters: Dict[str, Callable]) -> Dict[str, Dict[str, Callable]]:
    """Special types that must be explicitly converted after reading them with pd.read_csv"""
    converters = {}
    for tb in dg_column_types.keys():
        converters[tb] = {}
        for k, v in dg_column_types[tb].items():
            if v == "vector":
                converters[tb][k] = lambda x: np.array(ast.literal_eval(x), dtype=np.float32)
            if v == "datetime64[ns]":
                converters[tb][k] = pd.to_datetime
            if v == "object":
                if k not in object_converters:
                    logger.warning(f"Column '{k}' is object and has no object converter callback. Reading as string.")
                    continue
                converters[tb][k] = object_converters[k]
    return converters

def _read_column_types(path: Path, columns: Dict[str, List[str]]) -> Dict[str, Dict[str, ColumnType]]:
    """Reads dataset types from types.json expected in the same dir as the csv files"""
    col_types_file = path / "types.json"
    with open(col_types_file, "r", encoding="utf8") as json_file:
        raw_dtypes: Dict[str, Dict[str, str]] = json.load(json_file)

    col_types_file = {}
    for tb, tb_columns in columns.items():
        col_types_file[tb] = {}
        for column in tb_columns:
            assert column in raw_dtypes[tb], f"'{tb}/{column}' not in {raw_dtypes[tb]} (path: '{col_types_file}')"
            column_dtype = raw_dtypes[tb][column]
            assert column_dtype in VALID_COLUMN_TYPES, f"Got '{column_dtype}'. Valid: {VALID_COLUMN_TYPES}"
            col_types_file[tb][column] = column_dtype

    return col_types_file

def csv_peek(path: Path, tables_pk: Dict[str, str]) -> Dict[str, Dict[str, str]]:
    """Reads just the first row of all csv files given a path. Then, reads the types.json file and returns it."""
    cols = {}
    for tb in tables_pk.keys():
        cols[tb] = pd.read_csv(f"{path}/{tb}.csv", nrows=0).columns.tolist()
    res = _read_column_types(path, cols)
    return res

def csv_read(path: Path, tables_pk: Dict[str, str], columns: Dict[str, List[str]],
             object_converters: Optional[Dict[str, Callable]] = None) -> Tuple[Dict, Dict]:
    """Reads data from a path of CSVs, given a list of columns and a list of data-groups plus their primary keys"""
    assert path.is_dir(), f"Path must be a directory. Got '{path}'."
    assert isinstance(columns, dict), f"Expected dict, got {columns}"

    object_converters = object_converters if object_converters is not None else {}
    column_types = _read_column_types(path, columns)
    converters = _build_converters(column_types, object_converters)
    logger.info(f"Reading Dataset from CSV files. Path: '{path}'. Data groups: {list(tables_pk)}")

    data = {}
    for dg, pk in tables_pk.items():
        file_name = path / f"{dg}.csv"
        dtypes = {k: column_type_to_dtype(v) for k, v in column_types[dg].items() if k not in converters[dg]}
        dtypes = {k: v if v != pd.CategoricalDtype else "category" for k, v in dtypes.items()}
        # This is here so we can read w/o explicitly putting pks in the columns. They are deduced by being pks.
        logger.debug(f"Reading the CSV file: '{file_name}' ({len(columns[dg])} columns)")
        read_data = pd.read_csv(file_name, usecols=columns[dg], dtype=dtypes).set_index(pk)

        # force float32 so we don't get weird float64/float32 differences during computation
        float64_cols = read_data.select_dtypes(include="float64").columns
        read_data = read_data.astype({col_name: np.float32 for col_name in float64_cols})

        for col, converter_fn in converters[dg].items():
            logger.debug(f"Converting '{col}' with dtype '{column_types[dg][col]}'")
            if column_types[dg][col] == "datetime64[ns]":
                read_data[col] = converter_fn(read_data[col])
            else:
                read_data[col] = read_data[col].apply(converter_fn)

        data[dg] = read_data
    return data, column_types



"""Fundamental column types to Dataset library"""
# pylint: disable=protected-access
from typing import Dict, Optional
import numpy as np
import pandas as pd
from loguru import logger

ColumnType = str
# { users: {client_id: str, total_number_of_sessions: int}, sessions: {...}, ... }
DatasetDgTypes = Dict[str, Dict[str, ColumnType]]

VALID_COLUMN_TYPES = ["int", "float", "bool", "object", "str", "categorical", "vector",
                      "datetime64[ns]", "timedelta64[ns]"]

def column_type_to_dtype(column_type: ColumnType) -> type:
    """Converts a dataset type, represented as a string to a python dtype"""
    if column_type == "int":
        return int
    if column_type == "float":
        return float
    if column_type == "bool":
        return bool
    if column_type == "object":
        return object
    if column_type == "str":
        return str
    if column_type == "categorical":
        return pd.CategoricalDtype
    if column_type == "vector":
        return np.ndarray
    if column_type == "datetime64[ns]":
        return np.dtype("datetime64[ns]")
    if column_type == "timedelta64[ns]":
        return np.dtype("timedelta64[ns]")
    raise KeyError(f"Dataset type '{column_type}' not supported")

def _dtype_to_column_type(col_data: pd.Series, dtype: type) -> ColumnType:
    """Converts a regular dtype from a backend (i.e. pandas) to a dataet type, which is represented as string"""
    # This is first checked because it's not a numpy type, but a pandas type. All the other ones are numpy dtypes
    # used by pandas in their series.
    if isinstance(dtype, pd.CategoricalDtype):
        return "categorical"
    if isinstance(dtype, int) or np.issubdtype(dtype, np.dtype("int")):
        return "int"
    if isinstance(dtype, float) or np.issubdtype(dtype, np.dtype("float")):
        return "float"
    if isinstance(dtype, bool) or np.issubdtype(dtype, np.dtype("bool")):
        return "bool"
    if np.issubdtype(dtype, np.dtype("datetime64")):
        return "datetime64[ns]"
    if np.issubdtype(dtype, np.dtype("timedelta64")):
        return "timedelta64[ns]"
    if isinstance(dtype, object) or np.issubdtype(dtype, np.dtype("object")):
        if not isinstance(col_data.iloc[0], np.ndarray):
            return "object"
        if len(col_data) > 100_000: # TODO: add tests
            col_data = col_data.sample(n=100_000)
        return "object" if len(col_data.apply(len).unique()) > 1 else "vector"
    if isinstance(dtype, str) or np.issubdtype(dtype, np.dtype("str")):
        return "str"
    raise KeyError(f"Dtype '{dtype}' not supported")

def _check_col_data(col_data: pd.Series, col_data_type: ColumnType):
    """checks a single column given it's data type"""
    name = col_data.name
    if col_data_type == "vector":
        first_item = col_data.iloc[0]
        if not isinstance(first_item, np.ndarray):
            raise TypeError(f"Column '{name}' with data_type 'vector' has vectors stored as '{type(first_item)}'")
    if col_data_type == "categorical":
        if not isinstance(col_data.dtype, pd.CategoricalDtype):
            raise TypeError(f"Column '{name}' with data_type 'categorical' has dtype '{col_data.dtype}'")

def _auto_detect_column_types(dataset: "Dataset") -> DatasetDgTypes:
    logger.debug("Automatically detecting data types from pandas_dataset. It's recommended you provided this dict yourself.")
    res = {}
    for dg in dataset._data.keys():
        res[dg] = {}
        dg_data: pd.DataFrame = dataset._data[dg].reset_index()
        for col in dg_data:
            col_data = dg_data[col]
            res[dg][col] = _dtype_to_column_type(col_data, col_data.dtype)
    return res

def check_column_types(dataset: "Dataset", column_types: DatasetDgTypes) -> DatasetDgTypes:
    """Just checks the validity of the provided data types against the actual data in the dataset"""
    if column_types.keys() != dataset._data.keys():
        raise KeyError(f"Provided data types: {column_types.keys()}. Dataset: {dataset._data.keys()}")
    for dg in dataset._data.keys():
        for col in dataset._data[dg].columns:
            assert column_types[dg][col] in VALID_COLUMN_TYPES, f"{column_types[dg][col]} & {VALID_COLUMN_TYPES}"
            _check_col_data(dataset._data[dg][col], column_types[dg][col])
    return column_types

def build_column_types(dataset: "Dataset", column_types: Optional[DatasetDgTypes]) -> DatasetDgTypes:
    """Returns a properly verified dictionary of dg => {col: data type} for all data groups"""
    if column_types is None:
        column_types = _auto_detect_column_types(dataset)

    return check_column_types(dataset, column_types)



"""Statistics for the given dataset. Only applies for numerical columns, plus vetors and categoricals"""
from typing import Dict

import numpy as np
import pandas as pd

from loguru import logger

from .column_types import ColumnType

def compute_stats(dataset: "Dataset") -> Dict[str, Dict[str, np.ndarray]]:
    """computes the mins/maxs of all columns of all data groups and returns them as a dict"""
    res = {}
    for dg in dataset.keys():
        res[dg] = {}
        for col in dataset[dg].columns:
            res[dg][col] = _compute_col_stats(dataset[dg][col], dataset.column_types[dg][col])
    return res

def _compute_col_stats(col_data: pd.Series, col_data_type: ColumnType) -> np.ndarray:
    if col_data_type in ("object", "str", "datetime64[ns]", "timedelta64[ns]"):
        return None

    if col_data_type in ("float", "int"):
        return _numeric_col_stats(col_data, col_data_type)

    if col_data_type == "categorical":
        cats = col_data.dtype.categories
        return np.array([[0] * len(cats), [1] * len(cats)])

    if col_data_type == "bool":
        return np.array([0, 1])

    if col_data_type == "vector":
        n_cols = len(col_data.iloc[0])
        return np.array([[0] * n_cols, [1] * n_cols])

    raise TypeError(f"Unknown type reached. Column: '{col_data.name}'. Type: {col_data_type}")

def _numeric_col_stats(col_data: pd.Series, col_data_type: ColumnType) -> np.ndarray:
    assert col_data_type in ("float", "int")
    # Find the best min-max range from right to left. We're trying [5% - 95%] first, if it doesn't work, we'll try
    # [4% - 95%], then [4% - 96%], until [0.1% - 99.9%]. We never try real min/max as this could have outliers.
    percentiles_left = [0.1, 1, 2, 3, 4, 5]
    percentiles_right = [95, 96, 97, 98, 99, 99.9]
    potentials = np.percentile(col_data, [percentiles_left, percentiles_right])
    potential_min, potential_max = potentials
    if np.fabs(potential_min[0] - potential_max[-1]) < 1e-5:
        logger.warning(f"Column {col_data.name} has only one value: {potential_min[0]}.")
        return np.array([0, 1])

    i = len(potential_min) - 1 # start from right
    j = 0 # start from left
    k = 0 # odd/even counter
    while True:
        if i < 0 or j >= len(potential_min):
            break
        current_min = potential_min[i]
        current_max = potential_max[j]
        current_diff = np.fabs(current_min - current_max)
        # If we found a pair that has a value that is not identical (i.e. 5% percentile value == 95% percentile)
        if current_diff != 0:
            logger.debug(f"Column: '{col_data.name}. Min-max pair: [{percentiles_left[i]}% - {percentiles_right[j]}%]"
                         f" Values: {current_min:.2f} - {current_max:.2f}.")
            return np.array([current_min, current_max])
        k += 1
        i -= k % 2  # update only on odd iterations
        j += 1 - (k % 2)  # update only on even iterations

    raise ValueError(f"Could not compute a valid min-max for column '{col_data.name}'. "
                     f" Percentiles: {potential_min} and {potential_max}")



"""Writes a dataset to a parquet file"""
from pathlib import Path
import pandas as pd
import numpy as np

from loguru import logger

def parquet_write(dataset: "Dataset", path: Path):
    """Writes a parquet file to a path"""
    assert not path.is_file(), f"Path must be a directory. Got '{path}'."
    path.mkdir(exist_ok=True, parents=True)
    logger.info(f"Writing {dataset} as Parquet files. Path: '{path}'.")

    for dg in dataset.tables:
        output_path = path / f"{dg}.parquet"
        df: pd.DataFrame = dataset[dg]
        for col in df.columns:
            col_data = df[col]
            # convert to float32, as we have defined a rule in parquet_read to look only for float arrays
            if dataset.column_types[dg][col] == "vector":
                df = df.drop(columns=[col])
                df[col] = col_data.apply(lambda arr: arr.astype(np.float32))
            # list of lists with dtype object cannot be exported. If we shallow convert the outer array to list it's ok
            if dataset.column_types[dg][col] == "object" and isinstance(col_data.iloc[0], np.ndarray):
                df = df.drop(columns=[col])
                df[col] = col_data.apply(list)
        # reset index, because parquet reader is a bit weird and keeps the index as-is. It's simpler for now.
        df.to_parquet(output_path)



""""Merge module for dataset objects"""
from typing import Dict, Tuple
import pandas as pd
from .column_types import ColumnType

def merge_datasets(dataset1: "Dataset", dataset2: "Dataset", *args, must_be_disjoint: bool, **kwargs
                  ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Dict[str, ColumnType]]]:
    """Merges two datasets together. Calls underlying pd.merge only on needed data groups"""
    new_data: Dict[str, pd.DataFrame] = {}
    new_column_types: Dict[str, Dict[str, ColumnType]] = {}
    tbs_together = set(dataset1.tables).union(dataset2.tables)
    for tb in tbs_together:
        if tb not in dataset1.tables:
            new_data[tb] = dataset2[tb]
            new_column_types[tb] = dataset2.column_types[tb]
            continue

        if tb not in dataset2.tables:
            new_data[tb] = dataset1[tb]
            new_column_types[tb] = dataset1.column_types[tb]
            continue

        tb1: pd.DataFrame = dataset1[tb]
        tb2: pd.DataFrame = dataset2[tb]
        if tb2.index.name != tb1.index.name:
            raise KeyError(f"Index name mismatch: '{tb1.index.name}' vs '{tb2.index.name}'")

        common_cols = dataset1[tb].columns.intersection(tb2.columns)
        if len(common_cols) > 0 and must_be_disjoint:
            raise ValueError(f"Got common cols: {common_cols}, but must_be_disjoin is set to True.")

        tb2 = tb2.drop(columns=common_cols)
        new_data[tb] = pd.merge(tb1, tb2, on=dataset1.primary_keys[tb], *args, **kwargs)
        new_column_types[tb] = {**dataset1.column_types[tb], **dataset2.column_types[tb]}
    return new_data, new_column_types



"""Indexes internal module to compute the indexes of a dataset"""
from typing import Dict, List
import pandas as pd
import numpy as np

def compute_indexes(data: "Dataset") -> Dict[str, List]:
    """
    Computes a nested lists of indexes that helps to locate the data throughout a path:
    indexes from the `k`-th list help to locate the top node data points in the dataframe associated
    to the `k`-th node of the path.
    All the lists have the same length, equal to the length of the top node dataframe.

    Example:
        If the path is `users->sessions`, the `users` list contains just the row indexes oif the `users` dataframe
        (top node always has unique data points), and the `session` list contains the indexes from the `sessions`
        dataframe that correspond to users in the `users` dataframe.
        Thus, for each user in the top dataframe we know exactly where to find its sessions in the `sessions`
        dataframe.

    Returns a dict whose keys are the nodes in `path` and whose values are the correspondingly nested lists of
    indexes.
    """
    res: Dict[str, List[List[int]]] = {}
    for dg in data.data_groups:
        # we start with the top-level data group. Since it is unique, it will a simple mapping: [ [0], [1], .., [n] ]
        assert len(pd.unique(data[dg[0]].index)) == len(data[dg[0]]), f"Index of data group '{dg[0]}' is not unique"
        res[dg[0]] = [ [i] for i in range(len(data[dg[0]])) ]

        for i in range(len(dg)-1):
            parent: pd.DataFrame = data[dg[i]]
            child: pd.DataFrame = data[dg[i+1]]
            assert len(pd.unique(child.index)) == len(child), f"Index of data group '{dg[i+1]}' is not unique"
            # Example: groupby dg='sessions' by 'Client ID', or dg='hits' by 'Session ID'
            indices: Dict[str, np.ndarray] = child.groupby(parent.index.name).indices
            # turn the {index1: array([ix1]), index2: array([ix2])} into [ [ix1], [ix2] ]
            res[f"{dg[i]}-{dg[i+1]}"] = [x.tolist() for x in indices.values()]
    return res



"""Dataset class module."""
from __future__ import annotations
from typing import Dict, List, Optional, Union, Tuple, Callable, Any, overload
from copy import copy, deepcopy
from pathlib import Path
import json
from natsort import natsorted
import pandas as pd
import numpy as np

from loguru import logger

from .indexes import compute_indexes
from .stats import compute_stats
from .column_types import ColumnType, build_column_types, check_column_types
from .internal import check_data_groups_consistency, check_nans, build_data, build_data_groups, flatten_list, StrOrPath
from .merge import merge_datasets
from .input_output import csv_read, csv_write, csv_peek, parquet_read, parquet_write, parquet_peek

class Dataset:
    """
    A `Dataset` represents a nested group of dataframes and contains the underlying data as a dict
    with string keys and `pd.DataFrame` values and the nested lists of indexes
    which help locating the corresponding data throughout the path.
    Notes:
    - each df must have a primary key (index) and each child dataframe must have a common join key with its parent.
    - all the columns, even inside the nested structure, must be named uniquely. The only exceptions are primary
    keys, as they are replicated across all data groups. The library will throw exceptions if this is not respected.
    - you can't have cycles in the tables (users -> sessions -> users). Join keys must be unidirectional
    - you can't more than 1 child/parent: dg1 -> [dg2, dg3] or [dg1, dg2] -> dg3 is not allowed, only dg1->dg2->dg3
    - you can have disjoin data groups: [products]. [users, sessions], as long as they are not interleaved

    Example:
         Assume we have a `users-sessions` dataset (one user has many sessions).
         The underlying data contains the users and the sessions dataframes. The indexes dict
         contains two lists for users and sessions respectively. The users list contains exactly the
         index of the users dataframe. The sessions list has the same size with the users list and
         the i-th elements in the sessions list provides the indices of all the sessions associated to
         the i-th user in the users list (i.e. for each user in the users dataframe, we know precisely
         how to retrieve all its sessions from the sessions dataframe).

    Args:
        data: The underlying data as a dict of `pd.DataFrames` or `pd.Series`.
        column_types: Optional argument to enforce some ambiguous column types (i.e. categoricals being numbers)
        check_consistency: If True checks the consistency of all the join keys.

    Attributes:
        data: The underlying data as a dict of `pd.DataFrame` with the keys ordered from root to leaf.
        indexes: The lists of the dataset's nested indexes.
        join_keys: The join keys of the dataset as a dict.
        primary_keys: The primary keys of the dataset as a dict.
        columns: The columns of the dataset values as a dict.
        all_column_names: The list of all column names, primary keys, and join keys for all tables.
        hash: The hashes of the dataset values as a dict.
        shape: The shapes of the dataset values as a dict.
        data_groups: The dict of lists representing the nested data groups.
        tables: The list of tables from which data_groups are formed.
        keys: The keys of the dataset (used as compatibility with dict).
        values: The values of the dataset (used as compatibility with dict).
        items: The items of the dataset (used as compatibility with dict).
    """
    def __init__(self,
                 data: Dict[str, Union[pd.DataFrame, pd.Series]],
                 column_types: Optional[Dict[str, List[ColumnType]]] = None,
                 check_consistency: bool = False,
                 ):
        assert isinstance(data, dict), f"Datasets can only be constructed out of dicts of dataframes. Got {type(data)}"
        assert len(data.keys()) > 0, "No data groups were provided"
        self._data: Dict[str, pd.DataFrame] = build_data(data)
        self._column_types = build_column_types(self, column_types)
        self._data_groups: List[List[str]] = build_data_groups(self)
        self._indexes: Optional[Dict[str, List]] = None
        self._stats: Dict[str, Dict[str, Dict]] = None
        self._metadata: Dict[str, Any] = {}
        if check_consistency:
            self.check_consistency()

    @property
    def data(self):
        """Returns the underlying data of this dataset."""
        return self._data

    @property
    def column_types(self) -> Dict[str, Dict[str, ColumnType]]:
        """Returns the dataset's data types dictionary"""
        assert self._column_types is not None, "Data types is not constructed. Call the constructor properly."
        return {dg: self._column_types[dg] for dg in self.tables}

    @property
    def dtypes(self) -> Dict[str, Dict[str, ColumnType]]:
        """same as self.column_types"""
        return self.column_types

    @property
    def indexes(self) -> Dict[str, List]:
        """Returns the nested indexes of the dataset."""
        if self._indexes is None:
            self._indexes = compute_indexes(self)
        return self._indexes

    @property
    def stats(self) -> Dict[str, Dict[str, np.ndarray]]:
        """
        Returns the statistics (min, max) of each column in the dataset
        Return format:
        {
            "dg1": {
                "col1": np.array([
                    min_x, # can be also an array if dataset_type of col is categorical/vector
                    max_y
                ])
            }
        }
        """
        if self._stats is None:
            self._stats = compute_stats(self)
        return self._stats

    @property
    def metadata(self) -> Dict[str, Any]:
        """Returns the metadata. It is guaranteed that it can be converted to a json representation"""
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: Dict[str, Any]):
        assert isinstance(metadata, dict), f"Expected a dictionary metadata, got {metadata}"
        assert json.loads(json.dumps(metadata)) == metadata, f"Metadata doesn't seem JSON-able: {metadata}"
        self._metadata = metadata

    @property
    def join_keys(self) -> Dict[str, List[str]]:
        """
        Returns the join keys of the dataset as a dict {data_group: [join keys]}. We assume that each nested group
        contains the primary key of its parent.
        """
        res = {}
        for dg in self.data_groups:
            dg_keys = []
            for tb in dg:
                res[tb] = copy(dg_keys)
                dg_keys.append(self.primary_keys[tb])
        return res

    @property
    def primary_keys(self) -> Dict[str, str]:
        """Returns the primary keys of the dataset as a dict {table: primary key}."""
        return {dg: self._data[dg].index.name for dg in self.tables}

    @property
    def columns(self) -> Dict[str, List[str]]:
        """Returns the columns of the dataset as a dict {table: [columns]}."""
        return {dg: self._data[dg].columns.tolist() for dg in self.tables}

    @property
    def all_column_names(self) -> List[str]:
        """Returns the values of all column names + primary keys + join keys for all data groups."""
        all_cols = flatten_list(self.columns.values())
        all_keys = flatten_list(self.primary_keys.values())
        return list(set(all_cols + all_keys))

    @property
    def hash(self) -> Dict[str, pd.Series]:
        """Returns the pandas hash of the dataset as a dict {data_group: hash}."""
        res = {}
        for tb in self.tables:
            hashable_df = self[tb].select_dtypes(exclude=[object])
            res[tb] = pd.util.hash_pandas_object(hashable_df, index=True)
        return res

    @property
    def shape(self) -> Dict[str, Tuple[int, int]]:
        """Returns the shapes of the dataset as a dict {data_group: dataframe_shape}."""
        return {dg: tuple(self._data[dg].shape) for dg in self.tables}

    @property
    def data_groups(self) -> List[List[str]]:
        """
        Returns the data groups as a list of lists. It is assumed that the data groups are designed as a forest of
        linear trees (1 child only). So, we might have [ [products], [users, sessions] ] as two data groups. Each inner
        data group is guaranteed to be returned from root to leaf, so no [ [products], [sessions (X), users] ].
        However, since all data groups are equal at top level, we may get them in any order, so
        [ [users, sessions], [products] ] and [ [products], [users, sessions] ]
        are both fine, because only the data groups are shuffled.
        """
        return self._data_groups

    @property
    def tables(self) -> List[str]:
        """Returns all the tables as a flattened list"""
        return flatten_list(self.data_groups)

    def keys(self):
        """Returns the data groups of the dataset. Used as compatibility to dicts. No guarantees on dg order!"""
        return self._data.keys()

    def values(self):  # pragma: no cover
        """Returns the dataframes of the dataset. Used as compatibility to dicts. No guarantees on dg order!"""
        return self._data.values()

    def items(self):
        """Returns the keys-values of the dataset. Used as compatibility to dicts. No guarantees on dg order!"""
        return self._data.items()

    def copy(self) -> Dataset:
        """Returns a (deep) copy of this dataset"""
        return deepcopy(self)

    # pylint: disable=protected-access
    def sort(self, inplace: bool) -> Dataset:
        """Sorts the index of all data groups of this dataset."""
        logger.debug(f"Sorting this dataset {self} (inplace={inplace})")
        dataset: Dataset = self if inplace else self.copy()
        for dg in dataset.tables:
            # We need to sort from the first join key to the last, including the index.
            # so, for dg=hits (users>sessions>hits), we first sort by client id, then session id and lastly hit id
            tmp_dg = dataset._data[dg].reset_index()
            sort_keys = [*self.join_keys[dg], self.primary_keys[dg]]
            tmp_dg = tmp_dg.sort_values(sort_keys)
            dataset._data[dg] = tmp_dg.set_index(self.primary_keys[dg])
        return self

    def sort_columns(self):
        """Sorts the columns of the dataset by name, inplace"""
        for dg in self.tables:
            cols = self.columns[dg]
            sorted_cols = natsorted(cols)
            self._data[dg] = self._data[dg][sorted_cols]

    def get_column(self, col_name: str) -> Tuple[str, pd.Series]:
        """Gets the data of a column, if it exists in this dataset, alongside the data group it is part of"""
        assert isinstance(col_name, str), f"Expected a string, got {col_name}"
        for k, v in self.items():
            if col_name == v.index.name:  # pragma: no cover
                return k, v.index
            if col_name in v.columns:
                return k, v[col_name]
        raise KeyError(f"Column '{col_name}' not found in dataset.")

    def merge(self, other: Union[Dataset, Dict[str, Union[pd.DataFrame, pd.Series]]],
              *args, must_be_disjoint: bool=False, **kwargs) -> Dataset:
        """Given another dataset, merge using the underlying pandas.merge method for each data group."""
        if isinstance(other, dict):
            logger.warning("Merge was called with a dionary of dataframes. Converting to Dataset.")
            other = Dataset(other)
        assert isinstance(other, Dataset), f"Can only merge two datasets together. Got '{type(other)}'"
        new_data, new_column_types = merge_datasets(self, other, *args, must_be_disjoint=must_be_disjoint, **kwargs)
        res = Dataset(new_data, new_column_types)
        # save some computation time later on.
        if self._indexes is not None:
            res._indexes = self._indexes
        elif other._indexes is not None:
            res._indexes = other._indexes
        return res

    def apply_merge(self, fn: Callable) -> Dataset:
        """
        Applies the function on the dataset, then merges the result. Useful for chaining features.
        Example: For f=feature callback. dataset = dataset.merge(f(dataset)) => dataset = dataset.apply_merge(f)
        """
        return self.merge(fn(self))

    # pylint: disable=unused-argument
    def check_consistency(self, data_groups: bool = True, check_for_nans: bool = True, column_types: bool = True):
        """
        Checks called after mutable methods that work with the underlying dataframes directly as we cannot ensure that
        the dataframe has not changed leading to corrupted memory. We can however, check afterwards that the datset is
        still clean.
        """
        if data_groups:
            check_data_groups_consistency(self)
        if check_for_nans:
            check_nans(self)
        if column_types:
            check_column_types(self, self.column_types)

    # I/O operations

    @staticmethod
    def read(path: StrOrPath, data_format: str, tables_pk: Dict[str, str],
             columns: Dict[str, List[str]] = None, **kwargs) -> Dataset:
        """
        Loads the dataset from a path given a mapping {data group => primary key} for all data groups and a list
        of columns for each data-groups. If columns is not set, it will read all the columns.
        Supported formats: "csv", "parquet". If indexes are provided, they will be used to set the index of the data.

        Usage: Datset.read("/path/to/disk_dir", "csv", tables_pk={"users": "client_id"}, [columns=col]}

        Parameters:
        - path: The path to the dataset on disk.
        - data_format: The format of the dataset. Supported: "csv", "parquet".
        - tables_pk: A mapping {table => primary key} for all tables.
        - columns: A mapping {table => list of columns} for all tables. If None, all columns will be read.
        - kwargs: Additional arguments to the underlying reader.
        """
        assert data_format in ("csv", "parquet"), f"Unsupported data format: {data_format}"
        path = Path(path)
        if columns is None:
            columns = Dataset.peek(path, data_format, tables_pk)
            # dataset.peek returns a {table: {col: col_tye}} dict, we only need the keys for the readers
            columns = {k: list(v.keys()) for k, v in columns.items()}

        for table in columns:
            if tables_pk[table] not in columns[table]:
                logger.warning(f"Table '{table}' didn't have pk '{tables_pk[table]}' in the list of cols.")
                columns[table] = [*columns[table], tables_pk[table]]

        if data_format == "csv":
            data, dataset_types = csv_read(path, tables_pk, columns, **kwargs)
        if data_format == "parquet":
            data, dataset_types = parquet_read(path, tables_pk, columns, **kwargs)
        dataset = Dataset(data, dataset_types)

        if (path / "indexes.npy").exists():
            dataset._indexes = np.load(path / "indexes.npy", allow_pickle=True).item()
        if (path / "metadata.json").exists():
            with open(path / "metadata.json", "r", encoding="utf-8") as f:
                dataset.metadata = json.load(f)
        return dataset

    @staticmethod
    def peek(path: StrOrPath, data_format: str, tables_pk: Dict[str, str]) -> Dict[str, Dict[str, str]]:
        """
        Reads only the header and the dataset types from a path given a mapping {data group => primary key}.
        Supported formats: "csv", "parquet".

        Usage: Dataset.peek("/path/to/disk_dir", "parquet", {"users": "client_id", "sessions": "session_id"})
        """
        if data_format == "csv":
            return csv_peek(Path(path), tables_pk)
        if data_format == "parquet":
            return parquet_peek(Path(path), tables_pk)
        raise NotImplementedError(f"Unsupported data format: {data_format}")

    def write(self, path: StrOrPath, data_format: str, write_indexes: bool = False, write_metadata: bool = False,
              write_stats: bool = False):
        """
        Saves the dataset to a path on the disk given a data format.
        Usage: dataset.write("/path/to/disk_dir", "csv")

        Parameters:
        - path: The path where the dataset's tables will be saved
        - data_format: The format to use for saving the dataset. Supported formats: "csv", "parquet".
        - write_indexes: Whether to write the indexes of the dataframes or not.
        - write_metadata: Whether to write the metadata of the dataset or not.
        - write_stats: Whether to write the stats of the dataset or not.
        """
        assert data_format in ("csv", "parquet"), f"Unsupported data format: {data_format}"
        path = Path(path)
        if data_format == "csv":
            csv_write(self, path)
        if data_format == "parquet":
            parquet_write(self, path)

        if write_indexes:
            np.save(path / "indexes.npy", self.indexes)
        if write_metadata:
            with open(path / "metadata.json", "w", encoding="utf-8") as filename:
                json.dump(self.metadata, filename, indent=4)
        if write_stats:
            np.save(path / "dataset_stats.npy", self.stats)

    @overload
    def __getitem__(self, index: str) -> pd.DataFrame:
        ...

    @overload
    def __getitem__(self, index: Dict) -> Dataset:
        ...

    @overload
    def __getitem__(self, index: List) -> Dataset:
        ...

    def __getitem__(self, index: Union[Dict, List, str]) -> Union[Dataset, pd.DataFrame]:
        if isinstance(index, dict):
            res, res_dtypes = {}, {}
            # We can also index a dataset with a subset of columns {"dg": [cols], "dg2": [cols2]}
            for dg, v in index.items():
                assert isinstance(v, (tuple, set, list)), f"Expeted a list to index columns using a dict, got {v}"
                _v = sorted(list(set([*[(f.name if hasattr(f, "name") else f) for f in v], *self.join_keys[dg]])))
                res[dg] = self._data[dg][_v]
                res_dtypes[dg] = {v: self.column_types[dg][v] for v in self._data[dg].columns}
            return Dataset(res, res_dtypes)

        if isinstance(index, list):
            res, res_dtypes = {}, {}
            for dg in index:
                res[dg] = self._data[dg]
                res_dtypes[dg] = self.column_types[dg]
            return Dataset(res, res_dtypes)

        return self._data[index]

    def __getattribute__(self, name: str) -> Any:
        try:
            return super().__getattribute__(name)
        except AttributeError as ex:
            # this enables syntax like dataset.users if 'users' is a table in the dataset.
            # Alternative to dataset["users"], similarily to how it's done in pandas as well for columns.
            try:
                return self._data[name]
            except:
                raise AttributeError(ex)

    def __setitem__(self, item, value):
        raise ValueError("Setting values to data groups is not allowed. Use dataset.merge() or Dataset() constructor.")

    def __contains__(self, key):
        return key in self.tables

    def __str__(self) -> str:
        f_str = "Dataset:"
        for k in self.tables:
            f_str += f" '{k} ('{self._data[k].index.name}')':  {self._data[k].shape}"
        return f_str

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: Dataset) -> bool:
        """Two datasets are equal if they have the same keys and the same dataframe values."""
        if not isinstance(other, Dataset):
            return False

        # check key equality
        if self.tables != other.tables:
            return False

        # check shape equality
        if self.shape != other.shape:
            return False

        # check columns equality. They can be permuted, but we still count them
        if sorted(self.columns) != sorted(other.columns):
            return False

        # check data types
        for tb in self.tables:
            _other_tb = other[tb][self[tb].columns]
            if (self[tb].dtypes != _other_tb.dtypes).any():
                return False

        # check index equality
        for tb in self.tables:
            if (self[tb].index != other[tb].index).any():
                return False

        # finally, check all the values via hashing the rows
        self_hash = self.hash
        other_hash = other.hash
        for tb in self.tables:
            if (self_hash[tb] != other_hash[tb]).any():
                return False
        return True

    def __iter__(self):
        return iter(self._data)



"""DatasetHashCheck module"""

from loguru import logger
from .dataset import Dataset

class DatasetHashCheck:
    """
    Dataset consistency class. Can be used in a 'with statement context manager'. It checks that the Dataset object has
    not changed on the previously existing columns. This is used to allow new columns to be added in an 'append only'
    fashion while making sure that the old ones were not modified between the initial call and the exist from the
    context manager. Useful for creating new features and guaranteeing dataset consistency.
    """
    def __init__(self, dataset: Dataset):
        self.dataset = dataset

        self.hash_before = dataset.hash
        self.shape_before = self.dataset.shape

    def __call__(self):
        logger.debug(f"Running dataset hash check on {self.dataset}")
        hash_after = self.dataset.hash

        if self.hash_before.keys() != hash_after.keys():
            raise ValueError(f"Keys have changed. Before: {self.hash_before.keys()}. After: {hash_after.keys()}")

        for dg in self.hash_before.keys():
            self.hash_before[dg].sort_index(inplace=True)
            hash_after[dg].sort_index(inplace=True)
            if (self.hash_before[dg] != hash_after[dg]).any():
                raise ValueError(f"Hash changed between enter and exit on data group '{dg}'. "
                                 f"Shape before: {self.shape_before[dg]}. After: {self.dataset.shape[dg]}. "
                                 "Did you add alter the original dataset instead of a new dataset at compute step ?")

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        return self.__call__()



"""Writes a dataset to a (Typed) CSV file"""
from pathlib import Path
import json
import pandas as pd
import numpy as np
from loguru import logger

def _arr_to_tuple(x: np.ndarray) -> tuple:
    """converts (nested) np arrays to nested tuples"""
    if len(x) == 0:
        return ()
    if isinstance(x[0], (np.ndarray, list)):
        return tuple(_arr_to_tuple(y) for y in x)
    return tuple(x)

def csv_write(dataset: "Dataset", path: Path):
    """Writes a typed csv file to a path"""
    assert not path.is_file(), f"Path must be a directory. Got '{path}'."
    path.mkdir(exist_ok=True, parents=True)
    logger.info(f"Writing {dataset} as CSV files. Path: '{path}'.")

    for dg in dataset.tables:
        output_path = path / f"{dg}.csv"
        df: pd.DataFrame = dataset[dg]
        for col in df.columns:
            # we have to convert vector types from np arrays into tuples so they are properly stored as CSV
            col_data = df[col]
            if dataset.column_types[dg][col] == "vector":
                df = df.drop(columns=[col])
                df[col] = col_data.apply(tuple)
            if dataset.column_types[dg][col] == "object":
                if isinstance(col_data.iloc[0], np.ndarray):
                    df = df.drop(columns=[col])
                    df[col] = col_data.apply(_arr_to_tuple)
        df.to_csv(output_path, index=True)

    with open(f"{path}/types.json", "w", encoding="utf8") as f:
        f.write(json.dumps(dataset.column_types, indent=2))



"""Various checks done on the dataset to ensure its consistency. Called explicitly only for efficiency purposes"""
# pylint: disable=protected-access, consider-iterating-dictionary
from typing import Dict, Union, List, Iterable, T
from pathlib import Path
from copy import copy
from collections import Counter
import pandas as pd
import numpy as np

from loguru import logger

StrOrPath = Union[Path, str]

def get_project_root() -> Path:
    """gets the root of the project for tests"""
    return Path(__file__).absolute().parents[1]

def flatten_list(x: Iterable[T]) -> List[T]:
    """Flattens a list of lists. Also flattens np arrays of object type, tuples and sets."""
    if x is None or len(x) == 0:
        return []
    res = []
    for item in x:
        if isinstance(item, (tuple, list, set)) or (isinstance(item, np.ndarray) and item.dtype == object):
            res.extend(flatten_list(item))
        else:
            res.append(item)
    return res

def _check_one_path(dataset: "Dataset", path):
    # Do a .unique() on each join key and check that data frames are perfectly inner joinable
    for i in range(len(path) - 1):
        parent: pd.DataFrame = dataset[path[i]]
        child: pd.DataFrame = dataset[path[i + 1]]
        index_col = parent.index
        if index_col.name not in child.columns:
            raise KeyError(f"Join key '{index_col.name}' not in child data frame: '{path[i + 1]}' (path: '{path}')")
        unique_join_key = child[index_col.name].unique()
        intersection = index_col.intersection(unique_join_key)
        union = parent.index.union(unique_join_key)
        diff = list(union.difference(intersection))
        if len(intersection) != len(union):
            raise ValueError(f"Child data frame '{path[i + 1]}' does not inner join with parent '{path[i]}'."
                             f"Differences on join key '{index_col.name}': '{diff}'")

def check_data_groups_consistency(dataset: "Dataset"):
    """Check that all nested data groups are perfectly inner-joinable on the join keys. Will throw an error if true"""
    logger.debug("Checking data groups consistency")
    for path in dataset.data_groups:
        _check_one_path(dataset, path)

def check_nans(dataset: "Dataset"):
    """Checks if the dataset has nans. Will throw an error if true"""
    logger.debug("Checking for nans inside the dataset")
    for tb in dataset.tables:
        nans = dataset[tb].isna().any() if len(dataset[tb]) > 0 else []
        if sum(nans):
            raise ValueError(f"NaNs or Infinities were found in the table '{tb}': {nans}")

def build_data(data: Dict[str, Union[pd.Series, pd.DataFrame]]) -> Dict[str, pd.DataFrame]:
    """Returns a proper dictionary of dataframes. Does basic checks as well."""
    # Convert series to dataframes if the case
    df_data: Dict[str, pd.DataFrame] = {}
    all_cols = []
    pks = []
    for dg, v in data.items():
        if isinstance(v, pd.Series):
            assert v.name, f"The value of data group '{dg}' is a `pd.Series` without name."
            data[dg] = v.to_frame()

        if not isinstance(data[dg], pd.DataFrame):
            raise ValueError(f"The value of data group '{dg}' must be pd.Series or pd.Dataframe. Got: {type(v)}")

        if data[dg].index.name is None:
            logger.warning(f"Data group '{dg}' has empty index name. Defaulting to '{dg}'")
            data[dg].index.name = dg

        df_data[dg] = data[dg]
        all_cols.extend(df_data[dg].columns.tolist())
        pks.append(df_data[dg].index.name)

    all_cols_except_pks = [col for col in all_cols if col not in pks]
    duplicates = {k: v for k,v in Counter(all_cols_except_pks).items() if v > 1}
    if len(duplicates) > 0:
        raise ValueError(f"There are duplicate column names across tables: {duplicates}")

    return df_data

def build_data_groups(dataset: "Dataset") -> List[List[str]]:
    """
    Builds the nested data groups for all top-levels.
    Example: for users>sessions and products, we will get [[users, sessions], [products]]
    Cycles are ot allowed. Children with 2 parents are not allowed. One parent with 2 children are not allowed.
     - [tb1, tb2, tb3, tb1] -> wrong (rule 1)
     - tb1 => [tb2, tb3] -> wrong (rule 2)
     - [tb1, tb2] => tb3 -> wrong (rule 3)
     - [tb1, tb2, tb3] -> ok
    """
    pks = {tb: dataset._data[tb].index.name for tb in dataset._data}
    rels = {}
    tables = list(dataset._data.keys())
    for tb in tables:
        cols = set(dataset._data[tb].columns)
        rels[tb] = []
        for other_tb in tables:
            if pks[other_tb] in cols:
                rels[tb].append(other_tb)

    top_levels = [tb for tb in tables if len(rels[tb]) == 0]
    if len(top_levels) == 0:
        raise ValueError(f"No top levels found: {rels}. Possible cycles.")
    rels_orig = copy(rels)

    result = []
    for top_level in top_levels:
        this_dg = [top_level]
        current = top_level
        while True:
            del rels[current]
            found_one = False
            for tb in rels.keys():
                if current not in rels[tb]:
                    continue
                rels[tb].remove(current)
                # if we find one table that has only and only the primary key of current table, we continue this dg
                if len(rels[tb]) == 0:
                    if found_one:
                        raise ValueError(f"There may be cycles in your data tables: {rels_orig}")
                    this_dg.append(tb)
                    found_one = True
            if found_one is False:
                break
            current = this_dg[-1]
        result.append(this_dg)

    if len(flatten_list(result)) != len(rels_orig):
        raise ValueError(f"Not all tables were put into data groups. There may be cycles: {rels_orig}")

    result = sorted(result, key=lambda dg: dg[0])
    return result
